<?php

class GILSGitlab{
    private $curl = false;
    
    
    
    function getCommit($proj_path, $commit_h){
        global $config;
        
        $url = $config->server . "/api/v4/projects/" . urlencode($proj_path) . "/repository/commits/" . $commit_h;
        
        $r = $this->callAPI("GET", $url, false, $config->do_auth);
        $commit = json_decode($r['result']);
        
        if (is_object($commit) && str_contains($commit->message, "Not Found")){
            http_response_code(404);
            echo "Not found";
            die;
        }
        
        $commit->title = GILSUtils::replaceCommitRefsEtc($commit->title, $proj_path);
        $commit->message = GILSUtils::replaceCommitRefsEtc($commit->message, $proj_path);
        
        $url = $config->server . "/api/v4/projects/" . urlencode($proj_path) . "/repository/commits/" . $commit_h . "/diff";
        $r = $this->callAPI("GET", $url, false, $config->do_auth);
        $commit_diff = json_decode($r['result']);
        
        $url = $config->server . "/api/v4/projects/" . urlencode($proj_path) . "/repository/commits/" . $commit_h . "/comments";
        $r = $this->callAPI("GET", $url, false, $config->do_auth);
        $comments = json_decode($r['result']);
        
        
        $comments_arr = array();
        foreach ($comments as $comment){
            $comment->note = GILSUtils::replaceCommitRefsEtc($comment->note, $proj_path);
            $comments_arr[] = $comment;
        }
        
        $o = array(
            "commit" => $commit,
            "diff" => $commit_diff,
            "comments" => $comments_arr,
            "project" => $this->getProject($proj_path)
        );
        return $o;
    }
    
    
    function getCommits($proj_path){
        global $config;
        
        $url = $config->server . "/api/v4/projects/" . urlencode($proj_path) . "/repository/commits/";
        
        $r = $this->callAPI("GET", $url, false, $config->do_auth);
        
        $commits = json_decode($r['result']);
        $commit_arr = array();
        
        if (!is_array($commits) && property_exists($commits, "message") && str_contains($commits->message, "Not Found")){
            http_response_code(404);
            echo "Not found";
            die;
        }
        
        $x = 0;
        while (true){
            if (!$commits){
                break;
            }
            foreach ($commits as $commit) {
                $commit->title = GILSUtils::replaceCommitRefsEtc($commit->title, $proj_path);
                $commit->message = GILSUtils::replaceCommitRefsEtc($commit->message, $proj_path);
                $commit_arr[] = $commit;
            }
            
            if (array_key_exists("next", $r['links'])){
                $b = $this->callAPI("GET", GILSUtils::forceHTTPS($r['links']["next"]["link"]), false, $config->do_auth);
                $r = $b;
                $commits = json_decode($r['result']);
            }else{
                break;
            }
            $x++;
        }
        
        $o = array(
            "commits" => $commit_arr,
            "project" => $this->getProject($proj_path)
        );

        return $o;        
    }
    
    
    function getContainers($proj_path){
        global $config;
        
        $url = $config->server . "/api/v4/projects/" . urlencode($proj_path) . "/registry/repositories/?tags=true";
        $r = $this->callAPI("GET", $url, false, $config->do_auth);
        
        $containers = json_decode($r['result']);    
        //print_r($containers);
        
        if (is_object($containers) && $containers->message){
            $containers = array();
        }
        
        $o = array(
            "containers" => $containers,
            "project" => $this->getProject($proj_path)
        );

        return $o;           
    }

    function getMileStone($proj_path, $mid){
        global $config;
        
        
        $url = $config->server . "/api/v4/projects/" . urlencode($proj_path) . "/milestones/" . $mid;
        $r = $this->callAPI("GET", $url, false, $config->do_auth);
        $milestone = json_decode($r['result']);
        
        
        $o = array(
                "id" => $milestone->id,
                "title" => $milestone->title,
                "project" => $proj_path,
                "description" => $milestone->description,
                "state" => $milestone->state,
                "created" => $milestone->created_at,
                "updated" => $milestone->updated_at,
                "due" => $milestone->due_date,
                "started" => $milestone->start_date,
                "web_url" => $milestone->web_url,
                "project" => $this->getProject($proj_path)
             );
        
        
        $url = $config->server . "/api/v4/projects/" . urlencode($proj_path) . "/milestones/" . $mid . "/issues";
        $r = $this->callAPI("GET", $url, false, $config->do_auth);
        
        $issues = json_decode($r['result']);
        $issues_arr = array();
        
        $x = 0;
        while (true){
            if (!$issues){
                break;
            }
            foreach ($issues as $issue) {

                $p = array(
                    "title" => $issue->title,
                    "description" => GILSUtils::replaceCommitRefsEtc($issue->description, $proj_path),
                    "state" => $issue->state,
                    "created" => $issue->created_at,
                    "closed" => $issue->closed_at,
                    "closed_by" => $issue->closed_by,
                    "labels" => $issue->labels,
                    "milestone" => $issue->milestone,
                    "assignees" => $issue->assignees,
                    "author" => $issue->author,
                    "type" => $issue->issue_type,
                    "assignee" => $issue->assignee,
                    "gitlab_url" => $issue->web_url,
                    "gitlab_reference" =>  $proj_path . "#" . $issue->iid,
                    "short_reference" =>  "#" . $issue->iid
                );
                
                $issues_arr[] = $p;
            }
            
            if (array_key_exists("next", $r['links'])){
                $b = $this->callAPI("GET", GILSUtils::forceHTTPS($r['links']["next"]["link"]), false, $config->do_auth);
                $r = $b;
                $issues = json_decode($r['result']);
            }else{
                break;
            }
            $x++;
        }
        $o['issues'] = $issues_arr;
        return $o;
    }
    
    function getRelease($proj_path, $rid){
        global $config;
        
        
        $url = $config->server . "/api/v4/projects/" . urlencode($proj_path) . "/releases/" . $rid;
        $r = $this->callAPI("GET", $url, false, $config->do_auth);
        $release = json_decode($r['result']);
        $o = array(
                "release" => $release,
                "project" => $this->getProject($proj_path)
             );
        
        return $o;
    }
    
    
    function getProjectList(){
        global $config;

        $url = $config->server . "/api/v4/projects?sort=asc&order_by=name";
        $r = $this->callAPI("GET", $url, false, $config->do_auth);
        
        $projects = json_decode($r['result']);
        $project_arr = array();
        $x = 0;
        while (true){
            if (!$projects){
                break;
            }

            foreach ($projects as $project) {

                $p = array(
                    "name" => $project->path,
                    "namespaced_name" => $project->name_with_namespace,
                    "project_key" => $project->path_with_namespace,
                    "gitlab_url" => $project->web_url,
                    "description" => $project->description,
                    "avatar" => $project->avatar_url
                );
                
                $project_arr[] = $p;
            }
            
            if (array_key_exists("next", $r['links'])){
                $b = $this->callAPI("GET", GILSUtils::forceHTTPS($r['links']["next"]["link"], false, $config->do_auth));
                $r = $b;
                $projects = json_decode($r['result']);
            }else{
                break;
            }
            $x++;
        }
        return $project_arr;
    }
    

    function getProjectAndIssues($proj_path){
        //https://docs.gitlab.com/ee/api/projects.html#get-single-project
        
        global $config;   
        $o = array("project" => $this->getProject($proj_path));
        
        $url = $config->server . "/api/v4/projects/" . urlencode($proj_path) . "/issues?sort=asc&order_by=created_at";
        //echo $url;
        $r = $this->callAPI("GET", $url, false, $config->do_auth);

        if ($r['result'] == '{"message":"404 Project Not Found"}'){
            http_response_code(404);
            echo "Not found";
            die;
        }
        
        //print_r($r);
        
        $issues = json_decode($r['result']);
        $issues_arr = array();
        $x = 0;
        while (true){
            if (!$issues){
                break;
            }
            foreach ($issues as $issue) {

                $p = array(
                    "title" => $issue->title,
                    "description" => GILSUtils::replaceCommitRefsEtc($issue->description, $proj_path),
                    "state" => $issue->state,
                    "created" => $issue->created_at,
                    "closed" => $issue->closed_at,
                    "closed_by" => $issue->closed_by,
                    "labels" => $issue->labels,
                    "milestone" => $issue->milestone,
                    "assignees" => $issue->assignees,
                    "author" => $issue->author,
                    "type" => $issue->issue_type,
                    "assignee" => $issue->assignee,
                    "gitlab_url" => $issue->web_url,
                    "gitlab_reference" =>  $issue->references->full,
                    "short_reference" =>  $issue->references->short
                );
                
                $issues_arr[] = $p;
            }
            
            if (array_key_exists("next", $r['links'])){
                $b = $this->callAPI("GET", GILSUtils::forceHTTPS($r['links']["next"]["link"]), false, $config->do_auth);
                $r = $b;
                $issues = json_decode($r['result']);
            }else{
                break;
            }
            $x++;
        }
    
        $o["issues"] = $issues_arr;
        
        // Check whether the project has a wiki
        $url = $config->server . "/api/v4/projects/" . urlencode($proj_path) . "/wikis";
        $r = $this->callAPI("GET", $url, false, $config->do_auth);
        
        if ($r['result'] != '{"message":"403 Forbidden"}'){
            $o['has_wiki'] = true;
        }else{
            $o['has_wiki'] = false;
        }
        
        
        $c = $this->getContainers($proj_path);
        $o['has_containers'] = (count($c['containers']) > 0)? true : false;
        
        // Get milestones
        $url = $config->server . "/api/v4/projects/" . urlencode($proj_path) . "/milestones";
        $r = $this->callAPI("GET", $url, false, $config->do_auth);
        $o["milestones"] = json_decode($r['result']);
        
        $url = $config->server . "/api/v4/projects/" . urlencode($proj_path) . "/releases";
        $r = $this->callAPI("GET", $url, false, $config->do_auth);
        $o["releases"] = json_decode($r['result']);
        
        return $o;
    }
    
    function getProjectWikiPages($proj_path){
        global $config;
        $url = $config->server . "/api/v4/projects/" . urlencode($proj_path) . "/wikis";
        
        $r = $this->callAPI("GET", $url, false, $config->do_auth);

        if ($r['result'] == '{"message":"403 Forbidden"}'){
            http_response_code(404);
            echo "No exist";            
            die();
        }
        
        // Otherwise, crack on
        $pages = json_decode($r['result']);
        $pages_arr = array();
        while (true){
            if (!$pages){
                break;
            }
            foreach ($pages as $page) {               
                $pages_arr[] = $page;
            }
            
            if (array_key_exists("next", $r['links'])){
                $b = $this->callAPI("GET", GILSUtils::forceHTTPS($r['links']["next"]["link"]), false, $config->do_auth);
                $r = $b;
                $pages = json_decode($r['result']);
            }else{
                break;
            }
            $x++;
        }

        $o = array(
            "pages" => $pages_arr,
            "project" => $this->getProject($proj_path)
            );   
        return $o;
    }
    
    function getWikiPage($proj_path, $slug){
        global $config;
        $url = $config->server . "/api/v4/projects/" . urlencode($proj_path) . "/wikis/" . $slug;
        $r = $this->callAPI("GET", $url, false, $config->do_auth);

        if ($r['result'] == '{"message":"403 Forbidden"}'){
            echo "No exist";
            http_response_code(404);
            die();
        }
        // Otherwise, crack on
        $page = json_decode($r['result']);
        
        $page->gitlab_url = $config->server . "/" . $proj_path . "/-/wikis/" . $page->slug;

        $o = array(
            "page" => $page,
            "project" => $this->getProject($proj_path)
        );    
        return $o;
    
    }
    
    function getImage($path){
        global $config;
        $url = $config->server . $path;
        $r = $this->callAPI("GET", $url, false, $config->do_auth);
        //print_r($r);
        //echo $url;
        //die;
        header("content-type", $r['headers']["content-type"][0]);
        ini_set('default_mimetype', $r['headers']["content-type"][0]);
        
        if (isset($r['headers']["content-length"])){
            header("Content-length", $r['headers']["content-length"][0]);
        }

        echo $r['result'];
        die;
    }
    
    function getSnapShot($path){
        global $config;
        $url = $config->server . "/$path";
        $r = $this->callAPI("GET", $url, false, $config->do_auth);
        header("content-type", $r['headers']["content-type"][0]);
        ini_set('default_mimetype', $r['headers']["content-type"][0]);
        
        if (isset($r['headers']["content-length"])){
            header("Content-length", $r['headers']["content-length"][0]);
        }

        echo $r['result'];
        die;        
    }
    
    function getProject($proj_path){
        global $config;   
        $url = $config->server . "/api/v4/projects/" . urlencode($proj_path);
        
        $r = $this->callAPI("GET", $url, false, $config->do_auth);
        return json_decode($r['result']);
    }
    
    function getIssueAndComments($proj_path, $issue_ref){
    // TODO - need to add project info
        // /projects/:id/issues/:issue_iid
    
        global $config;
        $url = $config->server . "/api/v4/projects/" . urlencode($proj_path) . "/issues/" . $issue_ref;
        $r = $this->callAPI("GET", $url, false, $config->do_auth);
        //print_r($r);
        
        if ($r['result'] == '{"message":"404 Not found"}'){
            http_response_code(404);
            echo "Not found";
            die;
        }
        
        $issue = json_decode($r['result']);

        
        
        $p = array(
            "title" => $issue->title,
            "description" => GILSUtils::replaceCommitRefsEtc($issue->description, $proj_path),
            "state" => $issue->state,
            "created" => $issue->created_at,
            "closed" => $issue->closed_at,
            "closed_by" => $issue->closed_by,
            "labels" => $issue->labels,
            "milestone" => $issue->milestone,
            "assignees" => $issue->assignees,
            "author" => $issue->author,
            "type" => $issue->issue_type,
            "assignee" => $issue->assignee,
            "gitlab_url" => $issue->web_url,
            "gitlab_reference" =>  $issue->references->full,
            "short_reference" =>  $issue->references->short
        );        
        
        $o = array("issue" => $p, 
                    "comment_count" => 0,
                    "project" => $this->getProject($issue->project_id)
                    );
        
        // Now we need to get comments
        $url = GILSUtils::forceHTTPS($issue->_links->notes);
        $r = $this->callAPI("GET", $url."?sort=asc&order_by=created_at", false, $config->do_auth);
        //print_r($r);
        
        
        $comments= json_decode($r['result']);
        $comments_arr = array();
        $x = 0;
        while (true){
            if (!$comments){
                break;
            }
            foreach ($comments as $comment) {

                $body = GILSUtils::replaceCommitRefsEtc($comment->body, $proj_path);
                $body = str_replace("](/uploads", "](uploads", $body);
                
                $p = array(
                    "id" => $comment->id,
                    "body" => $body,
                    "author" => $comment->author,
                    "created" => $comment->created_at,
                    "update" => $comment->updated_at,
                    "system_update" => $comment->system,
                    "rowtype" => "comment",
                );
                
                if (!$comment->system){
                    $o["comment_count"]++;
                }else{
                    $p['rowtype'] = "statechange";
                }
                
                $comments_arr[] = $p;
            }
            
            if (array_key_exists("next", $r['links'])){
            echo $r['links']["next"]["link"];
                $b = $this->callAPI("GET", GILSUtils::forceHTTPS($r['links']["next"]["link"]), false, $config->do_auth);
                $r = $b;
                $comments = json_decode($r['result']);
            }else{
                break;
            }
            $x++;
        }
        
        $o["comments"] = $comments_arr;
        return $o;
        
    }
    
/*    
    function callAPI($method, $url, $data = false)
    {
        // https://stackoverflow.com/a/9802854
        $curl = curl_init();

        switch ($method)
        {
            case "POST":
                curl_setopt($curl, CURLOPT_POST, 1);

                if ($data)
                    curl_setopt($curl, CURLOPT_POSTFIELDS, $data);
                break;
            case "PUT":
                curl_setopt($curl, CURLOPT_PUT, 1);
                break;
            default:
                if ($data)
                    $url = sprintf("%s?%s", $url, http_build_query($data));
        }

        // Optional Authentication:
        curl_setopt($curl, CURLOPT_HTTPAUTH, CURLAUTH_BASIC);
        curl_setopt($curl, CURLOPT_USERPWD, "username:password");

        curl_setopt($curl, CURLOPT_URL, $url);
        curl_setopt($curl, CURLOPT_RETURNTRANSFER, 1);

        $result = curl_exec($curl);

        curl_close($curl);

        return $result;
    }    
    
*/


    function callAPI($method, $url, $data = false, $do_auth = true)
    {
        // based on https://stackoverflow.com/a/9802854
        
        $url = trim($url);

        $headers = [];
        global $config;

        if (!$this->curl){
            $this->curl = curl_init();
        }

        switch ($method)
        {
            case "POST":
                curl_setopt($this->curl, CURLOPT_POST, 1);

                if ($data)
                    curl_setopt($this->curl, CURLOPT_POSTFIELDS, $data);
                break;
            case "PUT":
                curl_setopt($this->curl, CURLOPT_PUT, 1);
                break;
            default:
                if ($data)
                    $url = sprintf("%s?%s", $url, http_build_query($data));
        }

        // Optional Authentication:
        //curl_setopt($curl, CURLOPT_HTTPAUTH, CURLAUTH_BASIC);
        //curl_setopt($curl, CURLOPT_USERPWD, "username:password");
        
        if ($do_auth){
            if ($config->access_token){
                curl_setopt($this->curl, CURLOPT_HTTPHEADER, array(
                    'Authorization: Bearer ' . $config->access_token
                ));        
            }
        }

        curl_setopt($this->curl, CURLOPT_URL, $url);
        curl_setopt($this->curl, CURLOPT_RETURNTRANSFER, 1);

        curl_setopt($this->curl, CURLOPT_HEADERFUNCTION,
        function($curl, $header) use (&$headers)
        {
            $len = strlen($header);
            $header = explode(':', $header, 2);
            if (count($header) < 2) // ignore invalid headers
            return $len;

            $headers[strtolower(trim($header[0]))][] = trim($header[1]);

            return $len;
        }
        );


        $result = curl_exec($this->curl);

        $links = array();
        if (array_key_exists("link", $headers)){
            $links = GILSUtils::parseLinkHeader($headers['link'][0]);
        }
        return array("result" => $result, "headers" => $headers, "links" => $links);
    }    
    

    
}
